#define _DEFAULT_SOURCE
#include <stdint.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "ipc.h"
#include "dist.h"

int send(void * self, local_id dst, const Message * msg) {
  if( self == NULL )
    return 1;

  my_info_t *me = (my_info_t *) self;
  if( me->dist_info == NULL )
    return 1;

  if( dst < 0 || dst > me->dist_info->x )
    return 1;

  uint8_t base = me->id * (me->dist_info->proccnt - 1); // base of block of pipes for this process
  uint8_t idx = 0;

  if( me->id < dst )
    idx = base + dst - 1;
  else
    idx = base + dst;

  int    wr_dscr = me->dist_info->pipes[idx][1];
  size_t wr_size = sizeof(msg->s_header) + msg->s_header.s_payload_len;

  if( write( wr_dscr, msg, wr_size ) != wr_size )
    return 1;

  return 0;
}

int send_multicast(void * self, const Message * msg) {
  if( self == NULL )
    return 1;

  my_info_t *me = (my_info_t *) self;
  if( me->dist_info == NULL )
    return 1;

  size_t wr_size;
  int wr_dscr;

  uint8_t base = me->id * (me->dist_info->proccnt - 1); // base of block of pipes for this process
  uint8_t i;

  for( i = 0; i < me->dist_info->proccnt - 1; i++) {
    wr_dscr = me->dist_info->pipes[ base + i][1];
    wr_size = sizeof(msg->s_header) + msg->s_header.s_payload_len;
    if( write( wr_dscr, msg, wr_size ) != wr_size )
      return 1;
  }

  return 0;
}

int receive(void * self, local_id from, Message * msg) {
  if( self == NULL )
    return 1;

  my_info_t *me = (my_info_t *) self;
  if( me->dist_info == NULL )
    return 1;

  if( from < 0 || from > me->dist_info->x )
    return 1;
  
  uint8_t base = from * (me->dist_info->proccnt - 1); 
  uint8_t idx = ( me->id > from ? base + me->id - 1 : base + me->id );
  int descriptor = me->dist_info->pipes[idx][0];

  char buf[MAX_MESSAGE_LEN];
  int size;

  while( 1 ) {
    if( ( size = read( descriptor, buf, sizeof( buf ) ) ) > 0 ) {
      memcpy( msg, buf, sizeof( *msg ) );
      return 0;
    }
    usleep(100000); // sleep 100 ms
  }
  return -1;
}

int receive_any(void * self, Message * msg) {
  if( self == NULL )
    return 1;

  my_info_t *me = (my_info_t *) self;
  if( me->dist_info == NULL )
    return 1;
  
  uint8_t i, j = 0;
  int *descriptors = (int *)malloc( sizeof(int) * (me->dist_info->proccnt - 1) );
  for(i = 0; i < me->dist_info->proccnt; i++) {
    // don't see at pipes as ( ID -> Y )
    if( me->id == i )
      continue;

    // base of block of pipes for this process
    uint8_t base = i * (me->dist_info->proccnt - 1); 
    
    // index number in block of pipes for ( X -> ID )
    uint8_t idx = ( me->id > i ? base + me->id - 1 : base + me->id );

    descriptors[j++] = me->dist_info->pipes[idx][0];
  }
  
  char buf[MAX_MESSAGE_LEN];
  int size;

  while( 1 ) {
    for( i = 0; i < j; i++ ) {
      if( ( size = read( descriptors[i], buf, sizeof( buf ) ) ) > 0 ) {
        memcpy( msg, buf, sizeof( *msg ) );
        free( descriptors );
        return 0;
      }
    }
    usleep(100000); // sleep 100 ms
  }
  
  free( descriptors );
  return 1;
}
